@@ -1,8 +1,8 @@
 /*
- * cmusfm - debug.h
- * Copyright (c) 2014 Arkadiusz Bokowy
+ * debug.h
+ * Copyright (c) 2014-2017 Arkadiusz Bokowy
  *
- * This file is a part of a cmusfm.
+ * This file is a part of cmusfm.
  *
  * cmusfm is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,16 +22,15 @@
 #define CMUSFM_DEBUG_H_
 
 #if HAVE_CONFIG_H
-#include "../config.h"
+# include "../config.h"
 #endif
 
 #include <stdio.h>
 
-
 #if DEBUG
-#define debug(M, ...) fprintf(stderr, "DEBUG %s:%d: " M "\n", __FILE__, __LINE__, ##__VA_ARGS__)
+# define debug(M, ARGS...) fprintf(stderr, "DEBUG %s:%d: " M "\n", __FILE__, __LINE__, ## ARGS)
 #else
-#define debug(M, ...) {}
+# define debug(M, ARGS...) do {} while (0)
 #endif
 
-#endif  /* CMUSFM_DEBUG_H_ */
+#endif
