@@ -1,8 +1,8 @@
 /*
- * cmusfm - libscrobbler2.c
- * Copyright (c) 2011-2017 Arkadiusz Bokowy
+ * libscrobbler2.c
+ * Copyright (c) 2011-2023 Arkadiusz Bokowy
  *
- * This file is a part of a cmusfm.
+ * This file is a part of cmusfm.
  *
  * cmusfm is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,10 +18,6 @@
  * see <http://www.gnu.org/licenses/>.
  */
 
-#if HAVE_CONFIG_H
-#include "../config.h"
-#endif
-
 #include "libscrobbler2.h"
 
 #include <ctype.h>
@@ -33,23 +29,36 @@
 #include "debug.h"
 #include "md5.c"
 
+/**
+ * Convenient macro for getting "on the stack" array size. */
+#define ARRAYSIZE(a) (sizeof(a) / sizeof(*(a)))
 
-static char *mem2hex(char *dest, const unsigned char *mem, size_t n);
+/**
+ * Type of the request value. */
+enum sb_request_data_type {
+	SB_REQUEST_DATA_TYPE_STRING,
+	SB_REQUEST_DATA_TYPE_NUMBER,
+};
 
+/**
+ * Request data name-value element. */
+struct sb_request_data {
+	char *name;
+	enum sb_request_data_type type;
+	union {
+		const char *s;
+		unsigned long n;
+	} value;
+};
 
-/* used as a buffer for GET/POST server response */
+/**
+ * Data structure for GET/POST server response. */
 struct sb_response_data {
 	char *data;
 	int size;
 };
 
-/* used for quick URL and signature creation process */
-struct sb_getpost_data {
-	char *name;
-	char data_format;
-	void *data;
-};
-
+static char *mem2hex(char *dest, const unsigned char *mem, size_t n);
 
 /* CURL write callback function. */
 static size_t sb_curl_write_callback(char *ptr, size_t size, size_t nmemb,
@@ -58,12 +67,15 @@ static size_t sb_curl_write_callback(char *ptr, size_t size, size_t nmemb,
 	struct sb_response_data *rd = (struct sb_response_data *)data;
 	size *= nmemb;
 
-	debug("read: size: %zu, body: %s", size, ptr);
+	debug("Read: size: %zu, body: %s", size, ptr);
 
 	/* XXX: Passing a zero bytes data to this callback is not en error,
 	 *      however memory allocation fail is. */
-	if (!size || (rd->data = realloc(rd->data, rd->size + size + 1)) == NULL)
+	void *tmp = rd->data;
+	if (!size || (rd->data = realloc(rd->data, rd->size + size + 1)) == NULL) {
+		free(tmp);
 		return 0;
+	}
 
 	memcpy(&rd->data[rd->size], ptr, size);
 	rd->size += size;
@@ -84,12 +96,22 @@ static CURL *sb_curl_init(CURLoption method, struct sb_response_data *response)
 	curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 15);
 	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 5);
 
-#if CURLOPT_PROTOCOLS
-	curl_easy_setopt(curl, CURLOPT_PROTOCOLS, CURLPROTO_HTTP);
+#if LIBCURL_VERSION_NUM >= 0x071101 /* 7.17.1 */
+	curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);
+	curl_easy_setopt(curl, CURLOPT_MAXREDIRS, 5);
+	curl_easy_setopt(curl, CURLOPT_POSTREDIR, CURL_REDIR_POST_ALL);
+#endif
+
+#if LIBCURL_VERSION_NUM >= 0x075500 /* 7.85.0 */
+	curl_easy_setopt(curl, CURLOPT_PROTOCOLS_STR, "http,https");
+	curl_easy_setopt(curl, CURLOPT_REDIR_PROTOCOLS_STR, "http,https");
+#elif LIBCURL_VERSION_NUM >= 0x071304 /* 7.19.4 */
+	curl_easy_setopt(curl, CURLOPT_PROTOCOLS, CURLPROTO_HTTP | CURLPROTO_HTTPS);
+	curl_easy_setopt(curl, CURLOPT_REDIR_PROTOCOLS, CURLPROTO_HTTP | CURLPROTO_HTTPS);
 #endif
-	curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
 
 	curl_easy_setopt(curl, method, 1);
+	curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
 
 	memset(response, 0, sizeof(*response));
 	curl_easy_setopt(curl, CURLOPT_WRITEDATA, response);
@@ -108,7 +130,7 @@ static void sb_curl_cleanup(CURL *curl, struct sb_response_data *response) {
 static scrobbler_status_t sb_check_response(struct sb_response_data *response,
 		int curl_status, scrobbler_session_t *sbs) {
 
-	debug("check: status: %d, body: %s", curl_status, response->data);
+	debug("Check: status: %d, body: %s", curl_status, response->data);
 
 	/* network transfer failure (curl error) */
 	if (curl_status != 0) {
@@ -116,6 +138,12 @@ static scrobbler_status_t sb_check_response(struct sb_response_data *response,
 		return sbs->status = SCROBBLER_STATUS_ERR_CURLPERF;
 	}
 
+	/* curl write callback was not called, something was mighty wrong... */
+	if (response->data == NULL) {
+		sbs->errornum = CURLE_GOT_NOTHING;
+		return sbs->status = SCROBBLER_STATUS_ERR_CURLPERF;
+	}
+
 	/* scrobbler service failure */
 	if (strstr(response->data, "<lfm status=\"ok\"") == NULL) {
 		char *error = strstr(response->data, "<error code=");
@@ -131,64 +159,112 @@ static scrobbler_status_t sb_check_response(struct sb_response_data *response,
 	return sbs->status = SCROBBLER_STATUS_OK;
 }
 
-/* Generate MD5 scrobbler API method signature. */
-static void sb_generate_method_signature(struct sb_getpost_data *sb_data,
-		int len, uint8_t secret[16], uint8_t sign[MD5_DIGEST_LENGTH]) {
+/**
+ * Generate MD5 signature for API call. */
+static void sb_generate_method_signature(
+		const scrobbler_session_t *sbs,
+		const struct sb_request_data *sb_data,
+		size_t sb_request_elements,
+		uint8_t sign[MD5_DIGEST_LENGTH]) {
 
 	char secret_hex[16 * 2 + 1];
-	char tmp_str[2048], format[8];
-	int x, offset;
+	char tmp_str[2048];
+	size_t i, offset;
 
-	mem2hex(secret_hex, secret, 16);
+	offset = 0;
+	for (i = 0; i < sb_request_elements; i++) {
 
-	for (x = offset = 0; x < len; x++) {
+		switch (sb_data[i].type) {
+		case SB_REQUEST_DATA_TYPE_NUMBER:
+			/* discard zero numeric values */
+			if (sb_data[i].value.n == 0)
+				continue;
+			offset += snprintf(&tmp_str[offset], sizeof(tmp_str) - offset,
+					"%s%lu", sb_data[i].name, sb_data[i].value.n);
+			break;
+		case SB_REQUEST_DATA_TYPE_STRING:
+			/* discard NULL string values */
+			if (sb_data[i].value.s == NULL)
+				continue;
+			offset += snprintf(&tmp_str[offset], sizeof(tmp_str) - offset,
+					"%s%s", sb_data[i].name, sb_data[i].value.s);
+			break;
+		}
 
-		/* it means that if numerical data is zero it is also discarded */
-		if (sb_data[x].data == NULL)
-			continue;
+		/* check for potential overflow */
+		if (offset > sizeof(tmp_str))
+			break;
 
-		sprintf(format, "%%s%%%c", sb_data[x].data_format);
-		offset += sprintf(tmp_str + offset, format, sb_data[x].name, sb_data[x].data);
 	}
 
-	debug("signature data: %s", tmp_str);
-	strcat(tmp_str, secret_hex);
+	if (offset <= sizeof(tmp_str) - sizeof(secret_hex))
+		strcpy(&tmp_str[offset], mem2hex(secret_hex, sbs->secret, 16));
 	MD5((unsigned char *)tmp_str, strlen(tmp_str), sign);
+
+#if DEBUG
+	char *tmp;
+	if ((tmp = strstr(tmp_str, sbs->session_key)) != NULL)
+		memset(tmp, 'x', strlen(sbs->session_key));
+	debug("Signature data: %s", tmp_str);
+#endif
+
 }
 
-/* Make curl GET/POST string (escape data). */
-static char *sb_make_curl_getpost_string(CURL *curl, char *str_buffer,
-		struct sb_getpost_data *sb_data, int len) {
+/**
+ * Make curl GET/POST request string. */
+static char *sb_make_curl_request_string(
+		const scrobbler_session_t *sbs,
+		const struct sb_request_data *sb_data,
+		size_t sb_request_elements,
+		char *dest,
+		size_t dest_size,
+		CURL *curl) {
 
-	char *escaped_data, format[8];
-	int x, offset;
+	char *escaped_data;
+	size_t i, offset;
 
-	for (x = offset = 0; x < len; x++) {
+	offset = 0;
+	for (i = 0; i < sb_request_elements; i++) {
 
-		/* it means that if numerical data is zero it is also discarded */
-		if (sb_data[x].data == NULL)
-			continue;
-
-		sprintf(format, "%%s=%%%c&", sb_data[x].data_format);
-
-		/* escape for string data */
-		if (sb_data[x].data_format == 's') {
-			escaped_data = curl_easy_escape(curl, sb_data[x].data, 0);
-			offset += sprintf(str_buffer + offset, format,
-					sb_data[x].name, escaped_data);
+		switch (sb_data[i].type) {
+		case SB_REQUEST_DATA_TYPE_NUMBER:
+			/* discard zero numeric values */
+			if (sb_data[i].value.n == 0)
+				continue;
+			offset += snprintf(&dest[offset], dest_size - offset,
+					"%s=%lu&", sb_data[i].name, sb_data[i].value.n);
+			break;
+		case SB_REQUEST_DATA_TYPE_STRING:
+			/* discard NULL string values */
+			if (sb_data[i].value.s == NULL)
+				continue;
+			/* escape for string data */
+			escaped_data = curl_easy_escape(curl, sb_data[i].value.s, 0);
+			offset += snprintf(&dest[offset], dest_size - offset,
+					"%s=%s&", sb_data[i].name, escaped_data);
 			curl_free(escaped_data);
+			break;
 		}
-		else
-			/* non-string content - no need for escaping */
-			offset += sprintf(str_buffer + offset, format,
-					sb_data[x].name, sb_data[x].data);
+
+		/* check for potential overflow */
+		if (offset > dest_size)
+			break;
 	}
 
 	/* strip '&' from the end of the string */
-	str_buffer[offset - 1] = 0;
-	debug("params: %s", str_buffer);
+	dest[offset - 1] = '\0';
 
-	return str_buffer;
+#if DEBUG
+	char tmp_str[2048] = {};
+	char *tmp = strncpy(tmp_str, dest, sizeof(tmp_str) - 1);
+	if ((tmp = strstr(tmp, sbs->session_key)) != NULL)
+		memset(tmp, 'x', strlen(sbs->session_key));
+	debug("Request: %s", tmp_str);
+#else
+	(void)sbs;
+#endif
+
+	return dest;
 }
 
 /* Scrobble a track. */
@@ -203,25 +279,25 @@ scrobbler_status_t scrobbler_scrobble(scrobbler_session_t *sbs,
 	struct sb_response_data response;
 
 	/* data in alphabetical order sorted by name field (except api_sig) */
-	struct sb_getpost_data sb_data[] = {
-		{"album", 's', sbt->album},
-		{"albumArtist", 's', sbt->album_artist},
-		{"api_key", 's', api_key_hex},
-		{"artist", 's', sbt->artist},
-		/* {"context", 's', NULL}, */
-		{"duration", 'd', (void *)(long)sbt->duration},
-		{"mbid", 's', sbt->mbid},
-		{"method", 's', "track.scrobble"},
-		{"sk", 's', sbs->session_key},
-		{"timestamp", 'd', (void *)sbt->timestamp},
-		{"track", 's', sbt->track},
-		{"trackNumber", 'd', (void *)(long)sbt->track_number},
-		/* {"streamId", 's', NULL}, */
-		{"api_sig", 's', sign_hex},
+	const struct sb_request_data sb_data[] = {
+		{ "album", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->album } },
+		{ "albumArtist", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->album_artist } },
+		{ "api_key", SB_REQUEST_DATA_TYPE_STRING, { .s = api_key_hex } },
+		{ "artist", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->artist } },
+		/* { "context", SB_REQUEST_DATA_TYPE_STRING }, */
+		{ "duration", SB_REQUEST_DATA_TYPE_NUMBER, { .n = sbt->duration } },
+		{ "mbid", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->mb_track_id } },
+		{ "method", SB_REQUEST_DATA_TYPE_STRING, { .s = "track.scrobble" } },
+		{ "sk", SB_REQUEST_DATA_TYPE_STRING, { .s = sbs->session_key } },
+		{ "timestamp", SB_REQUEST_DATA_TYPE_NUMBER, { .n = sbt->timestamp } },
+		{ "track", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->track } },
+		{ "trackNumber", SB_REQUEST_DATA_TYPE_NUMBER, { .n = sbt->track_number } },
+		/* { "streamId", SB_REQUEST_DATA_TYPE_STRING }, */
+		{ "api_sig", SB_REQUEST_DATA_TYPE_STRING, { .s = sign_hex } },
 	};
 
-	debug("scrobble: %ld", sbt->timestamp);
-	debug("payload: %s - %s (%s) - %d. %s (%ds)",
+	debug("Scrobble: %ld", sbt->timestamp);
+	debug("Payload: %s - %s (%s) - %d. %s (%ds)",
 			sbt->artist, sbt->album, sbt->album_artist,
 			sbt->track_number, sbt->track, sbt->duration);
 
@@ -234,16 +310,17 @@ scrobbler_status_t scrobbler_scrobble(scrobbler_session_t *sbs,
 	mem2hex(api_key_hex, sbs->api_key, sizeof(sbs->api_key));
 
 	/* make signature for track.scrobble API call */
-	sb_generate_method_signature(sb_data, 11, sbs->secret, sign);
+	sb_generate_method_signature(sbs, sb_data, ARRAYSIZE(sb_data) - 1, sign);
 	mem2hex(sign_hex, sign, sizeof(sign));
 
 	/* make track.scrobble POST request */
-	sb_make_curl_getpost_string(curl, post_data, sb_data, 12);
+	sb_make_curl_request_string(sbs, sb_data, ARRAYSIZE(sb_data),
+			post_data, sizeof(post_data), curl);
 	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_data);
-	curl_easy_setopt(curl, CURLOPT_URL, SCROBBLER_URL);
+	curl_easy_setopt(curl, CURLOPT_URL, sbs->api_url);
 
 	sb_check_response(&response, curl_easy_perform(curl), sbs);
-	debug("scrobble status: %d", sbs->status);
+	debug("Scrobble status: %d", sbs->status);
 
 	sb_curl_cleanup(curl, &response);
 	return sbs->status;
@@ -262,23 +339,23 @@ static scrobbler_status_t sb_update_now_playing(scrobbler_session_t *sbs,
 	struct sb_response_data response;
 
 	/* data in alphabetical order sorted by name field (except api_sig) */
-	struct sb_getpost_data sb_data[] = {
-		{"album", 's', sbt->album},
-		{"albumArtist", 's', sbt->album_artist},
-		{"api_key", 's', api_key_hex},
-		{"artist", 's', sbt->artist},
-		/* {"context", 's', NULL}, */
-		{"duration", 'd', (void *)(long)sbt->duration},
-		{"mbid", 's', sbt->mbid},
-		{"method", 's', "track.updateNowPlaying"},
-		{"sk", 's', sbs->session_key},
-		{"track", 's', sbt->track},
-		{"trackNumber", 'd', (void *)(long)sbt->track_number},
-		{"api_sig", 's', sign_hex},
+	const struct sb_request_data sb_data[] = {
+		{ "album", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->album } },
+		{ "albumArtist", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->album_artist } },
+		{ "api_key", SB_REQUEST_DATA_TYPE_STRING, { .s = api_key_hex } },
+		{ "artist", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->artist } },
+		/* { "context", SB_REQUEST_DATA_TYPE_STRING }, */
+		{ "duration", SB_REQUEST_DATA_TYPE_NUMBER, { .n = sbt->duration } },
+		{ "mbid", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->mb_track_id } },
+		{ "method", SB_REQUEST_DATA_TYPE_STRING, { .s = "track.updateNowPlaying" } },
+		{ "sk", SB_REQUEST_DATA_TYPE_STRING, { .s = sbs->session_key } },
+		{ "track", SB_REQUEST_DATA_TYPE_STRING, { .s = sbt->track } },
+		{ "trackNumber", SB_REQUEST_DATA_TYPE_NUMBER, { .n = sbt->track_number } },
+		{ "api_sig", SB_REQUEST_DATA_TYPE_STRING, { .s = sign_hex } },
 	};
 
-	debug("now playing: %ld", sbt->timestamp);
-	debug("payload: %s - %s (%s) - %d. %s (%ds)",
+	debug("Now playing: %ld", sbt->timestamp);
+	debug("Payload: %s - %s (%s) - %d. %s (%ds)",
 			sbt->artist, sbt->album, sbt->album_artist,
 			sbt->track_number, sbt->track, sbt->duration);
 
@@ -288,16 +365,17 @@ static scrobbler_status_t sb_update_now_playing(scrobbler_session_t *sbs,
 	mem2hex(api_key_hex, sbs->api_key, sizeof(sbs->api_key));
 
 	/* make signature for track.updateNowPlaying API call */
-	sb_generate_method_signature(sb_data, 10, sbs->secret, sign);
+	sb_generate_method_signature(sbs, sb_data, ARRAYSIZE(sb_data) - 1, sign);
 	mem2hex(sign_hex, sign, sizeof(sign));
 
 	/* make track.updateNowPlaying POST request */
-	sb_make_curl_getpost_string(curl, post_data, sb_data, 11);
+	sb_make_curl_request_string(sbs, sb_data, ARRAYSIZE(sb_data),
+			post_data, sizeof(post_data), curl);
 	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_data);
-	curl_easy_setopt(curl, CURLOPT_URL, SCROBBLER_URL);
+	curl_easy_setopt(curl, CURLOPT_URL, sbs->api_url);
 
 	sb_check_response(&response, curl_easy_perform(curl), sbs);
-	debug("now playing status: %d", sbs->status);
+	debug("Now playing status: %d", sbs->status);
 
 	sb_curl_cleanup(curl, &response);
 	return sbs->status;
@@ -306,7 +384,7 @@ static scrobbler_status_t sb_update_now_playing(scrobbler_session_t *sbs,
 /* Update "Now playing" notification. */
 scrobbler_status_t scrobbler_update_now_playing(scrobbler_session_t *sbs,
 		scrobbler_trackinfo_t *sbt) {
-	debug("now playing wrapper");
+	debug("Now playing wrapper");
 	if (sbt->artist == NULL || sbt->track == NULL)
 		return sbs->status = SCROBBLER_STATUS_ERR_TRACKINF;
 	return sb_update_now_playing(sbs, sbt);
@@ -315,20 +393,27 @@ scrobbler_status_t scrobbler_update_now_playing(scrobbler_session_t *sbs,
 /* Hard-codded method for validating session key. This approach uses the
  * updateNotify method call with invalid parameters as a test call. */
 scrobbler_status_t scrobbler_test_session_key(scrobbler_session_t *sbs) {
-	debug("session validation wrapper");
+	debug("Session validation wrapper");
 	scrobbler_trackinfo_t sbt = { .artist = "", .track = "" };
-	return sb_update_now_playing(sbs, &sbt);
+	sb_update_now_playing(sbs, &sbt);
+	/* Because we are using invalid parameters for session key validation,
+	 * service might actually return the "Invalid Parameters" error code.
+	 * However, it means that the session key itself is valid. */
+	if (sbs->status == SCROBBLER_STATUS_ERR_SCROBAPI &&
+			sbs->errornum == SCROBBLER_API_ERR_INVALID_PARAMS)
+		return SCROBBLER_STATUS_OK;
+	return sbs->status;
 }
 
-/* Get the session key. The memory block pointed by the str has to be big
- * enough to contain sizeof(session_key). */
-char *scrobbler_get_session_key_str(scrobbler_session_t *sbs, char *str) {
-	return strncpy(str, sbs->session_key, sizeof(sbs->session_key));
+/* Get the session key. */
+const char *scrobbler_get_session_key(scrobbler_session_t *sbs) {
+	return sbs->session_key;
 }
 
 /* Set the session key. */
-void scrobbler_set_session_key_str(scrobbler_session_t *sbs, const char *str) {
-	strncpy(sbs->session_key, str, sizeof(sbs->session_key));
+void scrobbler_set_session_key(scrobbler_session_t *sbs, const char *str) {
+	strncpy(sbs->session_key, str, sizeof(sbs->session_key) - 1);
+	sbs->session_key[sizeof(sbs->session_key) - 1] = '\0';
 }
 
 /* Perform scrobbler service authentication process. */
@@ -342,18 +427,21 @@ scrobbler_status_t scrobbler_authentication(scrobbler_session_t *sbs,
 	char sign_hex[sizeof(sign) * 2 + 1], token_hex[33];
 	char get_url[1024], *ptr;
 	struct sb_response_data response;
+	size_t len;
 
 	/* data in alphabetical order sorted by name field (except api_sig) */
-	struct sb_getpost_data sb_data_token[] = {
-		{"api_key", 's', api_key_hex},
-		{"method", 's', "auth.getToken"},
-		{"api_sig", 's', sign_hex},
+	const struct sb_request_data sb_data_token[] = {
+		{ "api_key", SB_REQUEST_DATA_TYPE_STRING, { .s = api_key_hex } },
+		{ "method", SB_REQUEST_DATA_TYPE_STRING, { .s = "auth.getToken" } },
+		{ "api_sig", SB_REQUEST_DATA_TYPE_STRING, { .s = sign_hex } },
 	};
-	struct sb_getpost_data sb_data_session[] = {
-		{"api_key", 's', api_key_hex},
-		{"method", 's', "auth.getSession"},
-		{"token", 's', token_hex},
-		{"api_sig", 's', sign_hex},
+
+	/* data in alphabetical order sorted by name field (except api_sig) */
+	const struct sb_request_data sb_data_session[] = {
+		{ "api_key", SB_REQUEST_DATA_TYPE_STRING, { .s = api_key_hex } },
+		{ "method", SB_REQUEST_DATA_TYPE_STRING, { .s = "auth.getSession" } },
+		{ "token", SB_REQUEST_DATA_TYPE_STRING, { .s = token_hex } },
+		{ "api_sig", SB_REQUEST_DATA_TYPE_STRING, { .s = sign_hex } },
 	};
 
 	if ((curl = sb_curl_init(CURLOPT_HTTPGET, &response)) == NULL)
@@ -362,12 +450,13 @@ scrobbler_status_t scrobbler_authentication(scrobbler_session_t *sbs,
 	mem2hex(api_key_hex, sbs->api_key, sizeof(sbs->api_key));
 
 	/* make signature for auth.getToken API call */
-	sb_generate_method_signature(sb_data_token, 2, sbs->secret, sign);
+	sb_generate_method_signature(sbs, sb_data_token, ARRAYSIZE(sb_data_token) - 1, sign);
 	mem2hex(sign_hex, sign, sizeof(sign));
 
 	/* make auth.getToken GET request */
-	strcpy(get_url, SCROBBLER_URL "?");
-	sb_make_curl_getpost_string(curl, get_url + strlen(get_url), sb_data_token, 3);
+	len = snprintf(get_url, sizeof(get_url), "%s?", sbs->api_url);
+	sb_make_curl_request_string(sbs, sb_data_token, ARRAYSIZE(sb_data_token),
+			get_url + len, sizeof(get_url) - len, curl);
 	curl_easy_setopt(curl, CURLOPT_URL, get_url);
 
 	status = sb_check_response(&response, curl_easy_perform(curl), sbs);
@@ -377,30 +466,31 @@ scrobbler_status_t scrobbler_authentication(scrobbler_session_t *sbs,
 	}
 
 	memcpy(token_hex, strstr(response.data, "<token>") + 7, 32);
-	token_hex[32] = 0;
+	token_hex[32] = '\0';
 
 	/* perform user authorization (callback function) */
-	sprintf(get_url, SCROBBLER_USERAUTH_URL "?api_key=%s&token=%s",
-			api_key_hex, token_hex);
+	sprintf(get_url, "%s?api_key=%s&token=%s",
+			sbs->auth_url, api_key_hex, token_hex);
 	if (callback(get_url) != 0) {
 		sb_curl_cleanup(curl, &response);
 		return sbs->status = SCROBBLER_STATUS_ERR_CALLBACK;
 	}
 
 	/* make signature for auth.getSession API call */
-	sb_generate_method_signature(sb_data_session, 3, sbs->secret, sign);
+	sb_generate_method_signature(sbs, sb_data_session, ARRAYSIZE(sb_data_session) - 1, sign);
 	mem2hex(sign_hex, sign, sizeof(sign));
 
 	/* reinitialize response buffer */
 	response.size = 0;
 
 	/* make auth.getSession GET request */
-	strcpy(get_url, SCROBBLER_URL "?");
-	sb_make_curl_getpost_string(curl, get_url + strlen(get_url), sb_data_session, 4);
+	len = snprintf(get_url, sizeof(get_url), "%s?", sbs->api_url);
+	sb_make_curl_request_string(sbs, sb_data_session, ARRAYSIZE(sb_data_session),
+			get_url + len, sizeof(get_url) - len, curl);
 	curl_easy_setopt(curl, CURLOPT_URL, get_url);
 
 	status = sb_check_response(&response, curl_easy_perform(curl), sbs);
-	debug("authentication status: %d", sbs->status);
+	debug("Authentication status: %d", sbs->status);
 	if (status != SCROBBLER_STATUS_OK) {
 		sb_curl_cleanup(curl, &response);
 		return status;
@@ -425,8 +515,8 @@ scrobbler_status_t scrobbler_authentication(scrobbler_session_t *sbs,
 /* Initialize scrobbler session. On success this function returns pointer
  * to allocated scrobbler_session structure, which must be freed by call
  * to scrobbler_free. On error NULL is returned. */
-scrobbler_session_t *scrobbler_initialize(uint8_t api_key[16],
-		uint8_t secret[16]) {
+scrobbler_session_t *scrobbler_initialize(const char *api_url,
+		const char *auth_url, uint8_t api_key[16], uint8_t secret[16]) {
 
 	scrobbler_session_t *sbs;
 
@@ -434,11 +524,13 @@ scrobbler_session_t *scrobbler_initialize(uint8_t api_key[16],
 	if ((sbs = calloc(1, sizeof(scrobbler_session_t))) == NULL)
 		return NULL;
 
-	if (curl_global_init(CURL_GLOBAL_NOTHING) != 0) {
+	if (curl_global_init(CURL_GLOBAL_DEFAULT) != 0) {
 		free(sbs);
 		return NULL;
 	}
 
+	strncpy(sbs->api_url, api_url, sizeof(sbs->api_url) - 1);
+	strncpy(sbs->auth_url, auth_url, sizeof(sbs->auth_url) - 1);
 	memcpy(sbs->api_key, api_key, sizeof(sbs->api_key));
 	memcpy(sbs->secret, secret, sizeof(sbs->secret));
 
@@ -460,46 +552,64 @@ const char *scrobbler_strerror(scrobbler_session_t *sbs) {
 	case SCROBBLER_STATUS_ERR_CURLPERF:
 		return curl_easy_strerror(sbs->errornum);
 	case SCROBBLER_STATUS_ERR_SCROBAPI:
-		switch (sbs->errornum) {
-		case 2:
+		switch ((scrobbler_api_error_t)sbs->errornum) {
+		case SCROBBLER_API_ERR_INVALID_SERVICE:
 			return "API: Invalid service";
-		case 3:
+		case SCROBBLER_API_ERR_INVALID_METHOD:
 			return "API: Invalid method";
-		case 4:
+		case SCROBBLER_API_ERR_AUTH_FAILED:
 			return "API: Authentication failed";
-		case 5:
+		case SCROBBLER_API_ERR_INVALID_FORMAT:
 			return "API: Invalid format";
-		case 6:
+		case SCROBBLER_API_ERR_INVALID_PARAMS:
 			return "API: Invalid parameters";
-		case 7:
+		case SCROBBLER_API_ERR_INVALID_RESOURCE:
 			return "API: Invalid resource";
-		case 8:
+		case SCROBBLER_API_ERR_OPERATION_FAILED:
 			return "API: Operation failed";
-		case 9:
+		case SCROBBLER_API_ERR_INVALID_SESSION_KEY:
 			return "API: Invalid session key";
-		case 10:
+		case SCROBBLER_API_ERR_INVALID_API_KEY:
 			return "API: Invalid API key";
-		case 11:
+		case SCROBBLER_API_ERR_SERVICE_OFFLINE:
 			return "API: Service temporarily unavailable";
-		case 13:
+		case SCROBBLER_API_ERR_SUBSCRIBERS_ONLY:
+			return "API: Subscribers only";
+		case SCROBBLER_API_ERR_INVALID_SIGNATURE:
 			return "API: Invalid signature";
-		case 14:
+		case SCROBBLER_API_ERR_UNAUTHORIZED_TOKEN:
 			return "API: Token not authorized";
-		case 15:
+		case SCROBBLER_API_ERR_ITEM_NOT_AVAILABLE:
 			return "API: Token has expired";
-		case 16:
-			return "API: Temporary error - please tray again";
-		case 26:
+		case SCROBBLER_API_ERR_SERVICE_UNAVAILABLE:
+			return "API: Service temporarily unavailable";
+		case SCROBBLER_API_ERR_LOGIN_REQUIRED:
+			return "API: Login required";
+		case SCROBBLER_API_ERR_TRIAL_EXPIRED:
+			return "API: Trial expired";
+		case SCROBBLER_API_ERR_NOT_ENOUGH_CONTENT:
+		case SCROBBLER_API_ERR_NOT_ENOUGH_MEMBERS:
+		case SCROBBLER_API_ERR_NOT_ENOUGH_FANS:
+		case SCROBBLER_API_ERR_NOT_ENOUGH_NEIGHBOURS:
+			return "API: Not enough content/members/etc.";
+		case SCROBBLER_API_ERR_RADIO_NO_PEAK:
+			return "API: Radio no peak usage";
+		case SCROBBLER_API_ERR_RADIO_NOT_FOUND:
+			return "API: Radio not found";
+		case SCROBBLER_API_ERR_API_KEY_SUSPENDED:
 			return "API: Suspended API key";
-		case 29:
+		case SCROBBLER_API_ERR_DEPRECATED:
+			return "API: Deprecated request";
+		case SCROBBLER_API_ERR_LIMIT_EXCEDED:
 			return "API: Rate limit exceeded";
-		default:
-			return "API: Unknown error";
 		}
+		return "API: Unknown error";
 	case SCROBBLER_STATUS_ERR_CALLBACK:
 		return "Authentication callback failure";
 	case SCROBBLER_STATUS_ERR_TRACKINF:
 		return "Missing required data for track";
+	default:
+		return "Unknown status";
 	}
 }
 
